from abc import ABC, abstractmethod
import itertools
import operator
from functools import reduce
import json


def _sparse_mapping(param_sets, sim_ids, sweep_id, save):
    """
    Mapping function for sparse sweeps (not a full `n`-dimensional array).

    Returns a dictionary mapping simulation IDs to the parameter set used for
    that simulation.
    """
    sim_id_mapping = dict(zip(sim_ids, param_sets))

    if save:
        sim_ids_filename = f'sim_ids_{sweep_id}.json'

        with open(sim_ids_filename, 'w') as sim_ids_file:
            json.dump(sim_id_mapping, sim_ids_file)

    return sim_id_mapping


class Sweep(ABC):
    """
    Abstract class for parameter sweep types.

    Sweeps must define an initialization using ``__init__``, the sweep length
    using ``__len__``, an iteration using ``elements``, and a type of mapping
    using ``mapping``.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def __len__(self):
        """Return the number of elements of the sweep."""
        pass

    @abstractmethod
    def elements(self):
        """
        Return the elements of the sweep.

        May be lazily evaluated, depending on the type of sweep.
        """
        pass

    @abstractmethod
    def mapping(self, sim_ids, sweep_id, save=True):
        """
        Return a mapping between the simulation IDs and the parameter sets.

        The mapping may be from simulation IDs to parameter sets or vice-versa,
        depending on what is convenient for the type of sweep.

        Parameters
        ----------
        sim_ids : list
            The simulation IDs that were assigned to each parameter set, in the
            same order as returned by ``elements``.
        sweep_id : str
            The sweep ID
        save : bool
            Whether to save the mapping to disk. The filename should be of the
            form ``sim_ids_{sweep_id}``, with an extension depending on the
            mapping type. True by default.

        """
        pass


class CartesianSweep(Sweep):
    """
    A Cartesian product parameter sweep.

    Parameters
    ----------
    sweep_params : dict
        A dictionary containing the parameter names as keys and lists of values
        to sweep over as values.

    """

    def __init__(self, sweep_params):
        self.keys = list(sweep_params.keys())
        self.values = list(sweep_params.values())

        self.lengths = [len(value) for value in self.values]

    def __len__(self):
        return reduce(operator.mul, self.lengths, 1)

    def elements(self):
        product = itertools.product(*self.values)

        return (dict(zip(self.keys, element)) for element in product)

    def mapping(self, sim_ids, sweep_id, save=True):
        """
        Return a labelled array which maps parameters to simulation IDs.

        See :func:`parasweep.sweepers.Sweep.mapping` for argument information.
        Returns a multidimensional labelled array (using xarray) which maps
        the parameters to the simulation IDs. The array coordinates correspond
        to each sweep parameter, while the values contain the simulation IDs.
        If ``save=True``, this array will be saved as a netCDF file with the
        name ``sim_ids_{sweep_id}.nc``.

        """
        import xarray
        import numpy

        sim_ids_array = xarray.DataArray(numpy.reshape(numpy.array(sim_ids),
                                                       self.lengths),
                                         coords=self.values, dims=self.keys,
                                         name='sim_id')

        if save:
            sim_ids_filename = f'sim_ids_{sweep_id}.nc'
            sim_ids_array.to_netcdf(sim_ids_filename)

        return sim_ids_array


class FilteredCartesianSweep(Sweep):
    """
    A parameter sweep which uses specified parameter sets.

    Parameters
    ----------
    sweep_params : dict
        A dictionary containing the parameter names as keys and lists of values
        to sweep over as values.
    filter_func : function
        A boolean function of parameter values; only parameter sets that
        return true will be included in the sweep. The arguments of the
        function should be named after the corresponding parameters. If not all
        the parameters in the sweep are used in the filtering, the ``**kwargs``
        argument should be defined, or else an error will be raised.

    """

    def __init__(self, sweep_params, filter_func):
        keys = list(sweep_params.keys())
        values = list(sweep_params.values())

        product = itertools.product(*values)

        # Here we cannot do lazy evaluation since we need the length of the
        # filtered elements.
        product_dicts = [dict(zip(keys, element)) for element in product]
        self.filtered = list(filter(lambda d: filter_func(**d), product_dicts))

    def __len__(self):
        return len(self.filtered)

    def elements(self):
        return self.filtered

    def mapping(self, sim_ids, sweep_id, save=True):
        """
        Return a dictionary which maps simulation IDs to parameter sets.

        See :func:`parasweep.sweepers.Sweep.mapping` for argument information.
        If ``save=True``, this dictionary will be saved as a JSON file with the
        name ``sim_ids_{sweep_id}.json``.
        """
        return _sparse_mapping(self.filtered, sim_ids, sweep_id, save)


class SetSweep(Sweep):
    """
    A Cartesian product parameter sweep with filtering.

    Parameters
    ----------
    param_sets : list
        A list containing the parameter sets to use in the sweep as
        dictionaries.

    """

    def __init__(self, param_sets):
        self.param_sets = param_sets

    def __len__(self):
        return len(self.param_sets)

    def elements(self):
        return self.param_sets

    def mapping(self, sim_ids, sweep_id, save=True):
        """
        Return a dictionary which maps simulation IDs to parameter sets.

        See :func:`parasweep.sweepers.Sweep.mapping` for argument information.
        If ``save=True``, this dictionary will be saved as a JSON file with the
        name ``sim_ids_{sweep_id}.json``.
        """
        return _sparse_mapping(self.param_sets, sim_ids, sweep_id, save)


class _RandomVariable(ABC):
    """
    Abstract class for random variables.

    This interface is modelled on SciPy's
    ``scipy.stats._distn_infrastructure.rv_generic`` generic random variable
    class. Random variables must implement an ``rvs``method to generate
    samples.
    """

    @abstractmethod
    def rvs(self, size, random_state=None):
        """
        Generate random samples.

        Parameters
        ----------
        size : int
            How many samples to draw
        random_state : numpy.random.RandomState instance, optional
            If provided, use the given random state in generating random
            numbers. Otherwise, use the global random state.

        """
        pass


class RandomSweep(Sweep):
    """
    A random parameter sweep.

    Each parameter is treated as an independent random variable with a given
    distribution.

    Parameters
    ----------
    sweep_params : dict
        A dictionary containing the parameter names as keys and SciPy random
        variables (i.e., instances of subclasses of ``_RandomVariable``, or of
        ``scipy.stats._distn_infrastructure.rv_generic``) as values.
    length : int
        The number of sets of random parameters to draw
    random_state : numpy.random.RandomState instance, optional
        If provided, will use the given random state in generating random
        numbers. By default, it uses the global random state.

    """

    def __init__(self, sweep_params, length, random_state=None):
        self.sweep_params = sweep_params
        self.length = length
        self.random_state = random_state

    def __len__(self):
        return self.length

    def elements(self):
        param_rvs = [rv.rvs(size=self.length, random_state=self.random_state)
                     for rv in self.sweep_params.values()]
        self.param_sets = [dict(param_set) for param_set in
                           zip(*(zip([key]*self.length, rvs) for key, rvs
                                 in zip(self.sweep_params.keys(),
                                        param_rvs)))]
        return self.param_sets

    def mapping(self, sim_ids, sweep_id, save=True):
        """
        Return a dictionary which maps simulation IDs to parameter sets.

        See :func:`parasweep.sweepers.Sweep.mapping` for argument information.
        If ``save=True``, this dictionary will be saved as a JSON file with the
        name ``sim_ids_{sweep_id}.json``.
        """
        return _sparse_mapping(self.param_sets, sim_ids, sweep_id, save)



"""Namers for generating simulation IDs."""
import math
from abc import ABC, abstractmethod
import json


class Namer(ABC):
    """
    Abstract class for assigning simulation IDs to simulation.

    Only the ``next`` method has to be implemented.
    """

    def start(self, length):
        """
        Initialize naming.

        Parameters
        ----------
        length : int
            Indicates how many total simulations are in the sweep.

        """
        pass

    @abstractmethod
    def generate_id(self, param_set, sweep_id):
        """Generate simulation ID for a given parameter set.

        Parameters
        ----------
        param_set : dict
            The parameter set
        sweep_id : str
            The sweep ID

        """
        pass


class SequentialNamer(Namer):
    """
    Name simulations with consecutive numbers and leading zeros.

    Parameters
    ----------
    zfill : None or int, optional
        If provided, sets the width to which the name string is to be
        padded with zeros.
    start_at : int, optional
        Sets the integer to start at in the sequential naming.

    Examples
    --------
    >>> counter = SequentialNamer()
    >>> counter.start(length=11)
    >>> counter.generate_id({'key1': 'value1'}, '')
    '00'
    >>> counter.generate_id({'key2': 'value2'}, '')
    '01'
    >>> counter.start(length=2)
    >>> counter.generate_id({'key1': 'value1'}, 'sweep_id')
    'sweep_id_0'

    """

    def __init__(self, zfill=None, start_at=0):
        self.zfill_arg = zfill
        self.start_at = start_at

    def start(self, length):
        self.count = self.start_at - 1

        # Need to have `zfill_arg` separate because otherwise state can persist
        # across multiple evaluations of `run_sweep`
        if self.zfill_arg is None:
            # Compute how many digits are needed to represent all the numbers
            self.zfill = (math.floor(math.log10(length - 1) + 1) if length != 1
                          else 1)
        else:
            self.zfill = self.zfill_arg
        self.length = length

    def generate_id(self, param_set, sweep_id):
        if self.count + 1 >= self.length + self.start_at:
            raise StopIteration
        self.count += 1
        if sweep_id:
            return '_'.join([sweep_id, str(self.count).zfill(self.zfill)])
        else:
            return str(self.count).zfill(self.zfill)


class HashNamer(Namer):
    """
    Name simulations using hashing.

    Parameters
    ----------
    hash_length : int, optional
        How many hexadecimal numbers to truncate the hash to. 6 by default.

    Examples
    --------
    >>> namer = HashNamer()
    >>> namer.generate_id({'key1': 'value1'}, '')
    '31fc462e'
    >>> namer.generate_id({'key2': 'value2'}, '')
    '9970c8f5'

    """

    def __init__(self, hash_length=8):
        self.hash_length = hash_length

    def generate_id(self, param_set, sweep_id):
        from hashlib import sha1

        h = sha1()
        h.update(bytes(json.dumps(param_set), 'utf-8'))
        h.update(bytes(sweep_id, 'utf-8'))

        return h.hexdigest()[:self.hash_length]


class SetNamer(Namer):
    """
    Name simulations consecutively with a provided iterable.

    Parameters
    ----------
    names : Iterable[str]
        The sequence of names to assign to consecutive simulations.

    Examples
    --------
    >>> namer = SetNamer(['name1', 'name2'])
    >>> namer.generate_id({'key1': 'value1'}, '')
    'name1'
    >>> namer.generate_id({'key2': 'value2'}, '')
    'name2'

    """

    def __init__(self, names):
        self.names = iter(names)

    def generate_id(self, param_set, sweep_id):
        return next(self.names)



"""Template engines for generating configuration files."""
from abc import ABC, abstractmethod
from string import Formatter


class TemplateEngine(ABC):
    """
    Abstract base class for template engines.

    Subclasses should implement the ``load`` and ``render`` methods. Make sure
    to implement errors for providing parameters not present in the template,
    and for using parameters in the template that are not provided.

    """

    @abstractmethod
    def load(self, paths):
        """
        Load configuration templates.

        Parameters
        ----------
        paths : list
            List of paths of configuration templates to load.

        """
        pass

    @abstractmethod
    def render(self, param_set):
        """
        Render a configuration file with the template engine.

        Parameters
        ----------
        param_set : dict
            Dictionary with parameters and their values.

        """
        pass


class PythonFormatTemplate(TemplateEngine):
    """Template engine using Python's string formatting."""

    def load(self, paths):
        self.templates = []
        for path in paths:
            with open(path, 'r', encoding='utf-8') as template_file:
                self.templates.append(template_file.read())

    def render(self, param_set):
        keys = param_set.keys()
        unused_names = set(keys)
        rendered = []
        for template in self.templates:
            config_names = set([elem[1] for elem in
                                Formatter().parse(template) if elem[1]])
            unused_names -= config_names
            try:
                rendered.append(template.format(**param_set))
            except KeyError as key:
                raise NameError(f'The name {key} is used in the template but '
                                'not provided.')
        if unused_names:
            raise NameError(f'The names {unused_names} are not used in the '
                            'template.')
        return rendered


def _mako_template_names(template):
    """
    Return all the used identifiers in the Mako template.

    From Igonato's code at https://stackoverflow.com/a/23577289/622408.
    """
    from mako import lexer, codegen

    lexer = lexer.Lexer(template)
    node = lexer.parse()
    # ^ The node is the root element for the parse tree.
    # The tree contains all the data from a template
    # needed for the code generation process

    # Dummy compiler. _Identifiers class requires one
    # but only interested in the reserved_names field
    compiler = lambda: None
    compiler.reserved_names = set()

    identifiers = codegen._Identifiers(compiler, node)
    return identifiers.undeclared


class MakoTemplate(TemplateEngine):
    """Template engine using Mako."""

    def load(self, paths):
        from mako.template import Template

        self.templates = []
        for path in paths:
            self.templates.append(Template(filename=path,
                                           input_encoding='utf-8',
                                           strict_undefined=True))

    def render(self, param_set):
        keys = param_set.keys()
        unused_names = set(keys)
        rendered = []
        for template in self.templates:
            config_names = _mako_template_names(template.source)
            unused_names -= config_names
            rendered.append(template.render_unicode(**param_set))
        if unused_names:
            raise NameError(f'The names {unused_names} are not used in the '
                            'template.')
        return rendered



"""
Test suite.

Some tests rely on drmaa, Mako, or xarray being installed.
"""
import unittest
import tempfile
import time
import warnings
import os
import json
import contextlib
import io

from parasweep import run_sweep
from parasweep import CartesianSweep, FilteredCartesianSweep, SetSweep, \
                      RandomSweep
from parasweep.namers import SequentialNamer, HashNamer, SetNamer
from parasweep.dispatchers import SubprocessDispatcher

# Use a Python "cat" and "sleep" command to make it more cross-platform
cat = ' '.join(['python', os.path.join(os.path.dirname(__file__), 'cat')])
sleep = ' '.join(['python',
                  os.path.join(os.path.dirname(__file__), 'sleep')])
sleepcat = ' '.join(['python',
                     os.path.join(os.path.dirname(__file__), 'sleepcat')])
err = ' '.join(['python', os.path.join(os.path.dirname(__file__), 'err')])


class TestSweep(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        warnings.simplefilter('ignore', ImportWarning)

    def test_basic(self):
        with tempfile.NamedTemporaryFile('r') as out, \
                tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x:.2f}\n')
            template.seek(0)

            run_sweep(' '.join([cat, ' {sim_id}.txt', out.name]),
                      ['{sim_id}.txt'], templates=[template.name],
                      sweep=CartesianSweep({'x': [1/3, 2/3, 3/3]}),
                      verbose=False, cleanup=True, save_mapping=False)

            # Check unordered, because we can't guarantee order
            self.assertEqual(set(out.read().splitlines()),
                             set(('Hello 0.33\nHello 0.67\n'
                                  'Hello 1.00\n').splitlines()))

    def test_single(self):
        with tempfile.NamedTemporaryFile('r') as out, \
                tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x}, {y}\n')
            template.seek(0)

            run_sweep(' '.join([cat, '{sim_id}.txt', out.name]),
                      ['{sim_id}.txt'], templates=[template.name],
                      sweep=CartesianSweep({'x': [1], 'y': [2]}),
                      verbose=False, cleanup=True, save_mapping=False)

            self.assertEqual(out.read(), 'Hello 1, 2\n')

    def test_multiple_templates(self):
        with tempfile.NamedTemporaryFile('r') as out, \
                tempfile.NamedTemporaryFile('w') as template1, \
                tempfile.NamedTemporaryFile('w') as template2:
            template1.write('Hello {x:.2f}, ')
            template1.seek(0)

            template2.write('hello again {y}\n')
            template2.seek(0)

            run_sweep(' '.join([cat, '{sim_id}_1.txt', '{sim_id}_2.txt',
                                out.name]),
                      ['{sim_id}_1.txt', '{sim_id}_2.txt'],
                      templates=[template1.name, template2.name],
                      sweep=CartesianSweep({'x': [1/3, 2/3, 3/3], 'y': [4]}),
                      verbose=False, cleanup=True, save_mapping=False)

            self.assertEqual(set(out.read().splitlines()),
                             set(('Hello 0.33, hello again 4\nHello 0.67, '
                                  'hello again 4\nHello 1.00, '
                                  'hello again 4\n').splitlines()))

    def test_mako(self):
        from parasweep.templates import MakoTemplate

        with tempfile.NamedTemporaryFile('r') as out, \
                tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello ${x*10}\n')
            template.seek(0)

            run_sweep(' '.join([cat, '{sim_id}.txt', out.name]),
                      ['{sim_id}.txt'], templates=[template.name],
                      sweep=CartesianSweep({'x': [1, 2, 3]}),
                      template_engine=MakoTemplate(), verbose=False,
                      cleanup=True, save_mapping=False)

            self.assertEqual(set(out.read().splitlines()),
                             set(('Hello 10\nHello 20\n'
                                  'Hello 30\n').splitlines()))

    def test_drmaa(self):
        from parasweep.dispatchers import DRMAADispatcher
        from drmaa import JobTemplate

        with tempfile.NamedTemporaryFile('r') as out, \
                tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x}\n')
            template.seek(0)

            run_sweep(' '.join([cat, '{sim_id}.txt', out.name]),
                      ['{sim_id}.txt'], templates=[template.name],
                      sweep=CartesianSweep({'x': [1, 1, 1]}), wait=True,
                      dispatcher=DRMAADispatcher(), verbose=False,
                      cleanup=True, save_mapping=False)

            self.assertEqual(out.read(), 'Hello 1\nHello 1\nHello 1\n')

            jt = JobTemplate(errorPath=':err_test.txt')
            run_sweep(' '.join([err, '{sim_id}']),
                      ['{sim_id}.txt'], templates=[template.name],
                      sweep=CartesianSweep({'x': [1]}), wait=True,
                      dispatcher=DRMAADispatcher(jt), verbose=False,
                      cleanup=True, save_mapping=False)

            with open('err_test.txt', 'r') as error_file:
                self.assertEqual('Error, simulation ID: 0\n',
                                 error_file.read())

            os.remove('err_test.txt')

    def test_serial(self):
        from parasweep.dispatchers import DRMAADispatcher

        with tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x}\n')
            template.seek(0)

            start_time = time.time()
            run_sweep(' '.join([sleep, str(2)]), ['{sim_id}.txt'],
                      templates=[template.name],
                      sweep=CartesianSweep({'x': [1, 2]}), serial=True,
                      verbose=False, save_mapping=False)

            self.assertGreater(time.time() - start_time, 4)

            start_time = time.time()
            run_sweep(' '.join([sleep, str(2)]), ['{sim_id}.txt'],
                      templates=[template.name],
                      sweep=CartesianSweep({'x': [1, 2]}),
                      dispatcher=DRMAADispatcher(), serial=True, verbose=False,
                      save_mapping=False)

            self.assertGreater(time.time() - start_time, 4)

        os.remove('0.txt')
        os.remove('1.txt')

    def test_wait(self):
        with tempfile.NamedTemporaryFile('r') as out, \
                tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x}\n')
            template.seek(0)

            run_sweep(' '.join([sleepcat, str(3), '{sim_id}.txt', out.name]),
                      ['{sim_id}.txt'], templates=[template.name],
                      sweep=CartesianSweep({'x': [1, 1, 1]}), wait=True,
                      verbose=False, save_mapping=False, sweep_id='test')

            self.assertEqual(out.read(), 'Hello 1\nHello 1\nHello 1\n')

        os.remove('test_0.txt')
        os.remove('test_1.txt')
        os.remove('test_2.txt')

    def test_errors(self):
        with tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x} {z}\n')
            template.seek(0)

            with self.assertRaises(TypeError) as context:
                run_sweep(' '.join([cat, '{sim_id}.txt']), '{sim_id}.txt',
                          templates=[template.name],
                          sweep=CartesianSweep({'x': [1, 2, 3]}),
                          verbose=False, cleanup=True, save_mapping=False)

            self.assertEqual('`configs` and `templates` must be a list.',
                             str(context.exception))

    def test_param_mapping(self):
        import xarray

        with tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x} {y} {z}\n')
            template.seek(0)

            param_array = run_sweep(' '.join([cat, '{sim_id}.txt']),
                                    ['{sim_id}.txt'],
                                    templates=[template.name],
                                    sweep=CartesianSweep({'x': [1, 2],
                                                          'y': [3, 4, 5],
                                                          'z': [6, 7, 8, 9]}),
                                    verbose=False, cleanup=True,
                                    sweep_id='test', save_mapping=True)

            self.assertEqual(param_array.coords.dims, ('x', 'y', 'z'))
            self.assertEqual(param_array.shape, (2, 3, 4))

            self.assertTrue(os.path.isfile('sim_ids_test.nc'))

            all_equal = (param_array
                         == xarray.open_dataarray('sim_ids_test.nc')).all()
            self.assertTrue(all_equal)

            os.remove('sim_ids_test.nc')

    def test_parameter_sets(self):
        with tempfile.NamedTemporaryFile('r') as out, \
                tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x}, {y}, {z}\n')
            template.seek(0)

            mapping = run_sweep(' '.join([cat, '{sim_id}.txt', out.name]),
                                ['{sim_id}.txt'],
                                templates=[template.name],
                                sweep=SetSweep([{'x': 2, 'y': 8, 'z': 5},
                                                {'x': 1, 'y': -4, 'z': 9}]),
                                verbose=False, cleanup=True,
                                sweep_id='test1', save_mapping=True)

            self.assertEqual(set(out.read().splitlines()),
                             set(('Hello 2, 8, 5\n'
                                  'Hello 1, -4, 9\n').splitlines()))
            self.assertEqual(mapping, {'test1_0': {'x': 2, 'y': 8, 'z': 5},
                                       'test1_1': {'x': 1, 'y': -4, 'z': 9}})

            self.assertTrue(os.path.isfile('sim_ids_test1.json'))

            with open('sim_ids_test1.json', 'r') as json_file:
                self.assertEqual(mapping, json.load(json_file))

            os.remove('sim_ids_test1.json')

    def test_filtered_cartesian(self):
        with tempfile.NamedTemporaryFile('r') as out, \
                tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x}, {y}\n')
            template.seek(0)

            sweep = FilteredCartesianSweep({'x': [1, 2, 3], 'y': [1, 2, 3]},
                                           filter_func=lambda x, y: x > y)
            mapping = run_sweep(' '.join([cat, '{sim_id}.txt', out.name]),
                                ['{sim_id}.txt'], templates=[template.name],
                                sweep=sweep, verbose=False, cleanup=True,
                                sweep_id='test2', save_mapping=True)

            self.assertEqual(set(out.read().splitlines()),
                             set(('Hello 2, 1\nHello 3, 1\n'
                                  'Hello 3, 2').splitlines()))

            self.assertTrue(os.path.isfile('sim_ids_test2.json'))

            with open('sim_ids_test2.json', 'r') as json_file:
                self.assertEqual(mapping, json.load(json_file))

            os.remove('sim_ids_test2.json')

    def test_random(self):
        import scipy.stats
        import numpy.random

        with tempfile.NamedTemporaryFile('r') as out, \
                tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x:.2f}, {y:.2f}\n')
            template.seek(0)

            sweep = RandomSweep({'x': scipy.stats.norm(),
                                 'y': scipy.stats.uniform(loc=0, scale=3)},
                                random_state=numpy.random.RandomState(30),
                                length=3)
            mapping = run_sweep(' '.join([cat, '{sim_id}.txt', out.name]),
                                ['{sim_id}.txt'], templates=[template.name],
                                sweep=sweep, verbose=False, cleanup=True,
                                sweep_id='test3', save_mapping=True)

            self.assertEqual(set(out.read().splitlines()),
                             set(('Hello -1.26, 2.89\nHello 1.53, 1.04\n'
                                  'Hello -0.97, 2.98').splitlines()))

            self.assertTrue(os.path.isfile('sim_ids_test3.json'))

            with open('sim_ids_test3.json', 'r') as json_file:
                self.assertEqual(mapping, json.load(json_file))

            os.remove('sim_ids_test3.json')

    def test_overwrite(self):
        with tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x}\n')
            template.seek(0)

            run_sweep(' '.join([cat, ' {sim_id}.txt']),
                      ['{sim_id}.txt'], templates=[template.name],
                      sweep=CartesianSweep({'x': [1]}),
                      verbose=False, cleanup=False, save_mapping=False,
                      sweep_id='test')

            with self.assertRaises(FileExistsError) as context:
                run_sweep(' '.join([cat, ' {sim_id}.txt']),
                          ['{sim_id}.txt'], templates=[template.name],
                          sweep=CartesianSweep({'x': [1]}),
                          verbose=False, cleanup=False, save_mapping=False,
                          overwrite=False, sweep_id='test')

            self.assertEqual('test_0.txt exists, set `overwrite` to True to '
                             'overwrite.', str(context.exception))

            os.remove('test_0.txt')

    def test_verbose(self):
        with tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x}\n')
            template.seek(0)

            temp_stdout = io.StringIO()
            with contextlib.redirect_stdout(temp_stdout):
                run_sweep(' '.join([cat, ' {sim_id}.txt']),
                          ['{sim_id}.txt'], templates=[template.name],
                          sweep=CartesianSweep({'x': [1]}),
                          verbose=True, cleanup=True, save_mapping=False, sweep_id='test')

            self.assertEqual(temp_stdout.getvalue(),
                             'Running simulation test_0 with parameters:\nx: 1\n')

    def test_max_procs(self):
        with tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x}\n')
            template.seek(0)

            dispatcher = SubprocessDispatcher(max_procs=1)

            start_time = time.time()
            run_sweep(' '.join([sleep, str(2)]), ['{sim_id}.txt'],
                      templates=[template.name],
                      sweep=CartesianSweep({'x': [1, 2]}),
                      dispatcher=dispatcher, verbose=False, save_mapping=False,
                      cleanup=True)

            self.assertGreater(time.time() - start_time, 4)


class TestPythonTemplates(unittest.TestCase):
    def test_errors(self):
        with tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello {x} {z}\n')
            template.seek(0)

            with self.assertRaises(NameError) as context:
                run_sweep(' '.join([cat, '{sim_id}.txt']), ['{sim_id}.txt'],
                          templates=[template.name],
                          sweep=CartesianSweep({'x': [1, 2, 3]}),
                          verbose=False, cleanup=True, save_mapping=False)

            self.assertEqual("The name 'z' is used in the template but not "
                             "provided.", str(context.exception))

            template.truncate(0)
            template.write('Hello {x}\n')
            template.seek(0)

            with self.assertRaises(NameError) as context:
                run_sweep(' '.join([cat, '{sim_id}.txt']), ['{sim_id}.txt'],
                          templates=[template.name],
                          sweep=CartesianSweep({'x': [1, 2, 3], 'y': [4]}),
                          verbose=False, cleanup=True, save_mapping=False)

            self.assertEqual("The names {'y'} are not used in the "
                             "template.", str(context.exception))


class TestMakoTemplates(unittest.TestCase):
    def test_errors(self):
        from parasweep.templates import MakoTemplate

        with tempfile.NamedTemporaryFile('w') as template:
            template.write('Hello ${x*10} ${z}\n')
            template.seek(0)

            with self.assertRaises(NameError) as context:
                run_sweep(' '.join([cat, '{sim_id}.txt']), ['{sim_id}.txt'],
                          templates=[template.name],
                          sweep=CartesianSweep({'x': [1, 2, 3]}),
                          template_engine=MakoTemplate(), verbose=False,
                          cleanup=True, save_mapping=False)

            self.assertEqual("'z' is not defined", str(context.exception))

            template.truncate(0)
            template.write('Hello ${x*10}\n')
            template.seek(0)

            with self.assertRaises(NameError) as context:
                run_sweep(' '.join([cat, '{sim_id}.txt']), ['{sim_id}.txt'],
                          templates=[template.name],
                          sweep=CartesianSweep({'x': [1, 2, 3], 'y': [4]}),
                          template_engine=MakoTemplate(), verbose=False,
                          cleanup=True, save_mapping=False)

            self.assertEqual("The names {'y'} are not used in the "
                             "template.", str(context.exception))


class TestNamers(unittest.TestCase):
    def test_sequential(self):
        counter = SequentialNamer()
        counter.start(length=11)

        self.assertEqual(counter.generate_id({'key1': 'value1'}, ''), '00')
        self.assertEqual(counter.generate_id({'key2': 'value2'}, ''), '01')

        counter = SequentialNamer(zfill=3, start_at=3)
        counter.start(length=2)

        self.assertEqual(counter.generate_id({'key1': 'value1'}, ''), '003')

        counter.generate_id({'key2': 'value2'}, '')

        with self.assertRaises(StopIteration):
            counter.generate_id({'key3': 'key_value3'}, '')

        counter = SequentialNamer()
        counter.start(length=1)

        self.assertEqual(counter.generate_id({'key1': 'value1'}, 'sweep_id'), 'sweep_id_0')

    def test_hash(self):
        counter = HashNamer()

        self.assertEqual(counter.generate_id({'key1': 'value1'}, ''),
                         '31fc462e')
        self.assertEqual(counter.generate_id({'key2': 'value2'}, ''),
                         '9970c8f5')

    def test_set(self):
        counter = SetNamer(['name1', 'name2'])

        self.assertEqual(counter.generate_id({'key1': 'value1'}, ''),
                         'name1')
        self.assertEqual(counter.generate_id({'key2': 'value2'}, ''),
                         'name2')



# -*- coding: utf-8 -*-
"""Main sweep functionality."""
import datetime
import os

from parasweep.namers import SequentialNamer
from parasweep.dispatchers import SubprocessDispatcher
from parasweep.templates import PythonFormatTemplate


def run_sweep(command, configs, templates, sweep, namer=SequentialNamer(),
              dispatcher=SubprocessDispatcher(),
              template_engine=PythonFormatTemplate(), sweep_id=None,
              excluded_sim_ids=[], serial=False, wait=False, cleanup=False,
              verbose=True, overwrite=True, save_mapping=True):
    r"""
    Run parameter sweeps.

    Parameters
    ----------
    command : str
        The command to run. Must include ``{sim_id}`` indicating where the
        simulation ID is to be inserted.
    configs : list
        List of paths indicating where the configuration files should be saved
        after substitution of the parameters into the templates. Each must
        include ``{sim_id}`` indicating where the simulation ID is to be
        inserted. Must be in the same order as ``templates``.
    templates : list
        List of paths of templates to substitute parameters into. Must be in
        the same order as ``configs``.
    sweep : sweepers.Sweep instance
        A :class:`parasweep.sweepers.Sweep` object that specifies the type of
        sweep. By default, it is a Cartesian sweep.
    namer : namers.Namer instance, optional
        A :class:`parasweep.namers.Namer` object that specifies how to assign
        simulation IDs. By default, assigns simulation IDs sequentially.
    dispatcher : dispatchers.Dispatcher instance, optional
        A :class:`parasweep.dispatchers.Dispatcher` object that specifies how
        to run the jobs. By default, uses Python's ``subprocess`` module.
    template_engine : templates.TemplateEngine instance, optional
        A :class:`parasweep.templates.TemplateEngine` object that specifies the
        template engine to use. By default, uses Python format strings.
    sweep_id : str, optional
        A name for the sweep. By default, the name is generated automatically
        from the date and time.
    excluded_sim_ids : list, optional
        A list of simulation IDs to exclude from the sweep.
    serial : bool, optional
        Whether to run simulations serially, i.e., to wait for each simulation
        to complete before executing the next one. Enabling this turns off
        parallelism. False by default.
    wait : bool, optional
        Whether to wait for all simulations to complete before returning.
        False by default.
    cleanup : bool, optional
        Whether to delete configuration files after all the simulations are
        done. This will cause the command to wait on all processes before
        returning (as with the ``wait`` argument). False by default.
    verbose : bool, optional
        Whether to print some information about each simulation as it is
        launched. True by default.
    overwrite : bool, optional
        Whether to overwrite existing files when creating configuration files.
        If False, a ``FileExistsError`` will be raised when a configuration
        filename coincides with an existing one in the same directory. True by
        default.
    save_mapping : bool, optional
        Whether to save a mapping between the parameters to the simulation
        IDs. The type of mapping will depend on the type of sweep (set with
        the ``sweep`` argument). The filename will be of the form
        ``sim_ids_{sweep_id}``, with an extension depending on the mapping
        type. True by default.

    Examples
    --------
    Many examples are provided in :doc:`examples`.

    """
    if isinstance(configs, str) or isinstance(templates, str):
        raise TypeError('`configs` and `templates` must be a list.')

    if sweep_id is None:
        current_time = datetime.datetime.now()
        sweep_id = current_time.strftime('%Y-%m-%dT%H_%M_%S')

    template_engine.load(paths=templates)

    namer.start(length=len(sweep))

    sim_ids = []
    config_filenames = []

    dispatcher.initialize_session()

    for param_set in sweep.elements():
        sim_id = namer.generate_id(param_set, sweep_id)
        sim_ids.append(sim_id)

        if sim_id not in excluded_sim_ids:
            rendered = template_engine.render(param_set)
            for config_rendered, config_path in zip(rendered, configs):
                config_filename = config_path.format(sim_id=sim_id)
                config_filenames.append(config_filename)
                if not overwrite:
                    if os.path.isfile(config_filename):
                        raise FileExistsError(f'{config_filename} exists, set '
                                              '`overwrite` to True to overwrite.')
                with open(config_filename, 'wb') as config_file:
                    config_file.write(config_rendered.encode('utf-8', 'replace'))

            if verbose:
                print(f'Running simulation {sim_id} with parameters:')
                print('\n'.join(f'{key}: {param}' for key, param
                                in param_set.items()))

            dispatcher.dispatch(command.format(sim_id=sim_id), serial)

    if wait:
        dispatcher.wait_all()

    if cleanup:
        dispatcher.wait_all()
        for config_filename in config_filenames:
            os.remove(config_filename)

    return sweep.mapping(sim_ids, sweep_id, save_mapping)
